@@ -91,12 +91,23 @@ int setsockopt(int fd, int level, int optname,
 # endif /* LINUX_VERSION_CODE */
 #endif /* __ARMEL__ */
 
+static gboolean gssdp_set_act_called , gssdp_set_act_val, prop_chd_emitted;
+
+static void ctrl_prop_chd(MafwExtension *ctrlsrc, gchar *name, GValue *nval,
+				gpointer udat)
+{
+	fail_if(strcmp(name, MAFW_PROPERTY_EXTENSION_ACTIVATE) != 0);
+	fail_if(G_VALUE_TYPE(nval) != G_TYPE_BOOLEAN);
+	prop_chd_emitted = TRUE;
+}
+
 START_TEST(test_plugin)
 {
 	MafwRegistry *reg = mafw_registry_get_instance();
 	gboolean retv;
 	GError *err = NULL;
 	GList *tlist;
+	MafwUpnpControlSource *ctrl_src;
 	
 	err = NULL;
 	retv = mafw_registry_load_plugin(reg,
@@ -107,10 +118,40 @@ START_TEST(test_plugin)
 	fail_unless(g_list_length(tlist) == 1);
 	g_list_free(tlist);
 	tlist = mafw_registry_get_sources(MAFW_REGISTRY(reg)); /* Do not free this list */
-	fail_unless(g_list_length(tlist) == 0);
+	fail_unless(g_list_length(tlist) == 1, "Src number: %d", g_list_length(tlist));
+	ctrl_src = MAFW_UPNP_CONTROL_SOURCE(tlist->data);
 	tlist = mafw_registry_get_renderers(MAFW_REGISTRY(reg)); /* Do not free this list */
 	fail_unless(g_list_length(tlist) == 0);
-	
+
+	fail_if(gssdp_set_act_called == TRUE);
+#ifndef HAVE_CONIC
+	g_signal_connect(G_OBJECT(ctrl_src), "property-changed",
+				(GCallback)ctrl_prop_chd, NULL);
+	mafw_extension_set_property_boolean(MAFW_EXTENSION(ctrl_src),
+					MAFW_PROPERTY_EXTENSION_ACTIVATE, TRUE);
+	fail_if(gssdp_set_act_called == FALSE);
+	fail_if(gssdp_set_act_val == FALSE);
+	fail_if(prop_chd_emitted == FALSE);
+	gssdp_set_act_called = FALSE;
+	mafw_extension_set_property_boolean(MAFW_EXTENSION(ctrl_src),
+					MAFW_PROPERTY_EXTENSION_ACTIVATE, FALSE);
+	/* it should not disable the browsing immediately */
+	fail_if(gssdp_set_act_called == TRUE);
+	fail_if(prop_chd_emitted == FALSE);
+	mafw_extension_set_property_boolean(MAFW_EXTENSION(ctrl_src),
+					MAFW_PROPERTY_EXTENSION_ACTIVATE, TRUE);
+	fail_if(gssdp_set_act_called == TRUE);
+	fail_if(prop_chd_emitted == FALSE);
+	checkmore_spin_loop(4000);
+	fail_if(gssdp_set_act_called == TRUE);
+	mafw_extension_set_property_boolean(MAFW_EXTENSION(ctrl_src),
+					MAFW_PROPERTY_EXTENSION_ACTIVATE, FALSE);
+	fail_if(gssdp_set_act_called == TRUE);
+	fail_if(prop_chd_emitted == FALSE);
+	checkmore_spin_loop(4000);
+	fail_if(gssdp_set_act_called == FALSE);
+	fail_if(gssdp_set_act_val == TRUE);
+#endif
 	mafw_registry_unload_plugin(reg, "mafw-upnp-source", &err);
 	fail_unless(retv && !err);
 	
@@ -178,7 +219,6 @@ void verify_results(struct expected_results *expected)
 
 static gboolean need_browse_results;
 static gboolean end_action_return_false;
-static gboolean with_wrong_didl;
 static gint browse_called;
 
 void browse_cb(MafwSource *source, guint browse_id, gint remaining,
@@ -187,7 +227,7 @@ void browse_cb(MafwSource *source, guint browse_id, gint remaining,
 {
 	/* Dummy browse result callback. */
 	browse_called++;
-	if (end_action_return_false || with_wrong_didl)
+	if (end_action_return_false)
 	{
 		fail_if(error == NULL);
 		fail_if(metadata != NULL);
@@ -223,7 +263,7 @@ static void mdata_result(MafwSource *self, const gchar *object_id,
 			const GError *error)
 {
 	mdata_called++;
-	if (end_action_return_false || with_wrong_didl)
+	if (end_action_return_false)
 	{
 		fail_if(error == NULL);
 		fail_if(metadata != NULL);
@@ -435,15 +475,6 @@ START_TEST(test_errors)
 	fail_if(browse_id == MAFW_SOURCE_INVALID_BROWSE_ID);
 	end_action_return_false = FALSE;
 
-	with_wrong_didl = TRUE;
-	browse_id = mafw_source_browse(source,
-				   "w::whatever", FALSE,
-				   NULL, NULL, MAFW_SOURCE_ALL_KEYS,
-				   0, 0,
-				   browse_cb, NULL);
-	fail_if(browse_id == MAFW_SOURCE_INVALID_BROWSE_ID);
-	with_wrong_didl = FALSE;
-
 	need_browse_results = FALSE;
 
 	return_null_action = TRUE;
@@ -464,16 +495,6 @@ START_TEST(test_errors)
 				 NULL);
 	end_action_return_false = FALSE;
 
-	with_wrong_didl = TRUE;
-	mafw_source_get_metadata(source,
-				 "w::whatever", 
-				 MAFW_SOURCE_ALL_KEYS,
-				 mdata_result,
-				 NULL);
-	with_wrong_didl = FALSE;
-
-	need_browse_results = FALSE;
-
 	mafw_upnp_source_plugin_deinitialize();
 	g_object_unref(source);
 }
@@ -671,6 +692,7 @@ int main(void)
 	tc = tcase_create("Init");
 	suite_add_tcase(suite, tc);
 if (1)	tcase_add_test(tc, test_plugin);
+	tcase_set_timeout(tc, 12);
 	
 	/* Browse tests */
 	tc = tcase_create("Browse");
@@ -789,14 +811,7 @@ gboolean gupnp_service_proxy_end_action(GUPnPServiceProxy *proxy,
 	(gchar *)va_arg(list, gchar*);
 	(gint)va_arg(list, gint);
 	data = va_arg(list, gpointer*);
-	if (with_wrong_didl)
-	{
-		*data = g_strdup(FAKE_DIDL_ITEM);
-	}
-	else
-	{
-		*data = g_strdup(DIDL_ITEM);
-	}
+	*data = g_strdup(DIDL_ITEM);
 	
 	if ((gchar *)va_arg(list, gchar*))
 	{
@@ -819,4 +834,11 @@ void gupnp_service_proxy_cancel_action (GUPnPServiceProxy *proxy,
 	return;
 }
 
+void gssdp_resource_browser_set_active(GSSDPResourceBrowser *resource_browser,
+                                                         gboolean active)
+{
+	gssdp_set_act_called = TRUE;
+	gssdp_set_act_val = active;
+}
+
 /* vi: set noexpandtab ts=8 sw=8 cino=t0,(0: */
